#!/usr/bin/env python3

import argparse
from typing import Optional
from dataclasses import dataclass
from .merger import Merger


version = "0.1.1"


@dataclass
class SubtitleConfig:
    """Configuration for a single subtitle file."""

    file_path: str
    encoding: Optional[str] = None
    e: Optional[str] = None
    color: Optional[str] = None
    c: Optional[str] = None
    top: bool = False


def create_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser."""
    parser = argparse.ArgumentParser(
        prog="SrtMerger",
        description="Merge multiple subtitle files with custom configurations.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument("-o", "--output", help="Output path")
    parser.add_argument(
        "-e", "--output-encoding", default="UTF8", help="Output encoding"
    )
    parser.add_argument(
        "-E",
        "--default-encoding",
        # default="UTF8",
        help="Default encoding for all files",
    )
    parser.add_argument("-C", "--default-color", help="Default color for all files")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {version}",
        help="Print the version",
    )

    parser.add_argument(
        "subtitles",
        nargs="+",
        metavar="FILE[:OPTS]",
        help="""Subtitle files with optional parameters.
                       Format: file.srt[:e|encoding=value][:c|color=value][:top].
                       Example: sub.srt:encoding=utf-8:color=yellow:top""",
    )

    return parser


def parse_subtitle_spec(spec: str) -> SubtitleConfig:
    """Parse a subtitle specification string into a SubtitleConfig.

    Format: filename.srt[:e|encoding=value][:c|color=value][:top]
    """
    parts = spec.split(":")
    file_path = parts[0]

    config = {"file_path": file_path}

    for part in parts[1:]:
        if part == "top":
            config["top"] = True
        elif "=" in part:
            key, value = part.split("=", 1)
            if key in ("encoding", "e", "color", "c"):
                config[key] = value

    return SubtitleConfig(**config)


def main():
    parser = create_parser()
    args = parser.parse_args()

    try:
        configs = [parse_subtitle_spec(spec) for spec in args.subtitles]
        merger = Merger(output_path=args.output, output_encoding=args.output_encoding)

        for config in configs:
            merger.add(
                config.file_path,
                encoding=config.encoding or config.e or args.default_encoding,
                color=config.color or config.c or args.default_color,
                top=config.top,
            )

        merger.merge()

    except Exception as e:
        parser.error(str(e))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import unittest
import os
from pathlib import Path

from srtmerger.merger import Merger


class TestSrtMerger(unittest.TestCase):
    test_assets_dir = Path(__file__).parent / "test_assets"
    filename = test_assets_dir / "test_ok.srt"

    def test_merge(self):
        merger = Merger(output_path=self.filename, output_encoding="utf-16-le")
        merger.add(
            self.test_assets_dir / "en.srt"
        )  # utf8 as default codec, white as default color
        merger.add(
            self.test_assets_dir / "fa.srt", color="yellow", encoding="cp1256", top=True
        )
        merger.merge()
        self.assertTrue(os.path.isfile(self.filename))

    def test_output_content(self):
        with open(self.filename, encoding="utf-16-le") as f:
            lines = f.readlines()
            self.assertTrue(os.path.isfile(self.filename))
            self.assertEqual(len(lines), 4232)
            self.assertEqual(lines[3091], "00:33:43,828 --> 00:33:46,796\n")
            self.assertEqual(lines[4228], "00:51:49,381 --> 00:51:59,286\n")

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.filename):
            os.remove(cls.filename)
            print(f'\n"{cls.filename}" was removed successfully.')


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Author: Iraj Jelodari

import logging
import codecs
import re
import enum
from pathlib import Path

from .utils import time_to_timestamp

# \r (Carriage Return) → moves the cursor to the beginning of the line without advancing to the next line
CR = "\r"

# \n (Line Feed) → moves the cursor down to the next line without returning to the beginning of the line
# — In a *nix environment \n moves to the beginning of the line.
LF = "\n"

LF_AS_BYTE = b"\n"

EOL = "\r\n"  # \r\n (End Of Line) → a combination of \r and \n

TIMECODES_PATTERN = (
    r"\d{1,2}:\d{1,2}:\d{1,2},\d{1,5} --> \d{1,2}:\d{1,2}:\d{1,2},\d{1,5}[\r\n]*"
)

logging.basicConfig(format="")

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Color(enum.StrEnum):
    RED = "#FF0000"
    DARK_RED = "#8B0000"
    CRIMSON = "#DC143C"
    ORANGE = "#FFA500"
    DARK_ORANGE = "#FF8C00"
    ORANGE_RED = "#FF4500"
    YELLOW = "#FFFF00"
    GOLD = "#FFD700"
    LIGHT_YELLOW = "#FFFFE0"
    GREEN = "#008000"
    LIME = "#00FF00"
    LIGHT_GREEN = "#90EE90"
    CYAN = "#00FFFF"
    DARK_CYAN = "#008B8B"
    LIGHT_CYAN = "#E0FFFF"
    BLUE = "#0000FF"
    DARK_BLUE = "#00008B"
    LIGHT_BLUE = "#ADD8E6"
    WHITE = "#FFFFFF"
    GRAY = "#808080"
    BLACK = "#000000"


class SrtParser:
    """
    A parser for SRT (SubRip) subtitle files.

    This class provides utilities to parse and manipulate SRT subtitle files.
    SRT files contain subtitle entries with the following structure:
    - Sequential Number: An integer indicating the subtitle's order
    - Timecodes: Start and end times in format HH:MM:SS,mmm --> HH:MM:SS,mmm
    - Subtitle Text: The actual text content (can span multiple lines)
    - Blank Line: Separates individual subtitle entries

    Methods:
        parse_timecodes: Extracts and returns the beginning and ending times from a timecode line
        parse_subtitle_entry: Parses a complete subtitle entry into its components (number, timecodes, and text)
        split_subtitle: Splits raw subtitle content by separating sequential numbers and timecode lines
    """

    def parse_timecodes(self, content):
        """Parse timecodes lines into begining time and end time"""
        PATTERN = r"(?P<begin>\d{1,2}:\d{1,2}:\d{1,2},\d{1,5})\s-->\s(?P<end>\d{1,2}:\d{1,2}:\d{1,2},\d{1,5})"
        match = re.search(PATTERN, content, re.MULTILINE | re.DOTALL | re.VERBOSE)
        return match.group("begin"), match.group("end")

    def parse_subtitle_entry(self, content):
        """Parse a subtitle entry (containing sequential number, timecodes and text) into sequential number, timecodes and text"""
        PATTERN = (
            r"(\d{1,2}:\d{1,2}:\d{1,2},\d{1,5}\s-->\s\d{1,2}:\d{1,2}:\d{1,2},\d{1,5})"
        )
        return re.split(PATTERN, content, flags=re.MULTILINE | re.DOTALL | re.VERBOSE)

    def split_subtitle(self, content):
        """Split a subtitle content by its sequential number and timecodes lines"""
        PATTERN = r"(\d*[\r\n]*\d{1,2}:\d{1,2}:\d{1,2},\d{1,5}\s-->\s\d{1,2}:\d{1,2}:\d{1,2},\d{1,5})"
        return re.split(PATTERN, content, flags=re.MULTILINE | re.DOTALL | re.VERBOSE)


class Merger:
    """
    SRT Merger allows you to merge subtitle files, no matter what language
    are the subtitles encoded in. The result of this merge will be a new subtitle
    file which will display subtitles from each merged file.

    Example:
    ```
    m = Merger(output_path="en-fa.srt")
    m.add(Path('./en.srt'), color=Color.BLUE)
    m.add(Path('./fa.srt'), color="yellow", encoding="cp1256")
    m.merge()
    ```
    """

    def __init__(
        self,
        output_path: str | Path = None,
        output_encoding: str = "utf-8",
    ):
        self.output_path = (
            Path(output_path) if type(output_path) is str else output_path
        )
        self.output_encoding = output_encoding
        self.parser = SrtParser()
        self.merged_subtitles = []
        self.timestamps = []
        self.subtitles = []

    def _insert_bom(self, content, encoding):
        encoding = encoding.replace("-", "").replace("_", "").replace(" ", "").upper()
        if encoding in ["UTF64LE", "UTF16", "UTF16LE"]:
            return codecs.BOM + content
        if encoding in ["UTF8"]:
            return codecs.BOM_UTF8 + content
        if encoding in ["UTF32LE"]:
            return codecs.BOM_UTF32_LE + content
        if encoding in ["UTF64BE"]:
            return codecs.BOM_UTF64_BE + content
        if encoding in ["UTF16BE"]:
            return codecs.BOM_UTF32_BE + content
        if encoding in ["UTF32BE"]:
            return codecs.BOM_UTF32_BE + content
        if encoding in ["UTF32"]:
            return codecs.BOM_UTF32 + content
        return content

    def _set_subtitle_color(self, text, color):
        """
        Set a color for subtitle
        """
        return f'<font color="{color}">{text}</font>' if color else text

    def _put_subtitle_top(self, subtitle):
        """
        Put the subtitle at the top of the screen
        """
        return "{\\an8}" + subtitle

    def _extract_subtitle_entries(self, data, color=None, top=False):
        subtitle_entries = dict()
        subtitle_list = self.parser.split_subtitle(data)[1:]

        # Items with even index represent sequential number and timecodes lines and
        # Items with odd index represent subtitle texts
        for seq_number_and_timecodes_idx in range(0, len(subtitle_list), 2):
            subtitle_entry = (
                subtitle_list[seq_number_and_timecodes_idx]
                + LF
                + subtitle_list[seq_number_and_timecodes_idx + 1].strip()
            )
            seq_number, timecodes, text = self.parser.parse_subtitle_entry(
                subtitle_entry
            )
            begin_time, end_time = self.parser.parse_timecodes(timecodes)

            logger.debug(f"index      = {seq_number_and_timecodes_idx}")
            logger.debug(f"seq_number = {seq_number.strip()}")
            logger.debug(f"timecodes  = {timecodes.strip()}")
            logger.debug(f"begin_time = {begin_time.strip()}")
            logger.debug(f"end_time   = {end_time.strip()}")
            logger.debug(f"text       = \n{text.strip()}")
            logger.debug(self.parser.parse_subtitle_entry(subtitle_entry))
            logger.debug("=" * 30)

            timestamp = time_to_timestamp(begin_time.strip())
            subtitle_text = self._set_subtitle_color(text.strip(), color)
            if top is True:
                subtitle_text = self._put_subtitle_top(subtitle_text)

            subtitle_entries[timestamp] = f"{timecodes.strip()}\n{subtitle_text}\n"

            self.timestamps.append(timestamp)
        return subtitle_entries

    def add(
        self,
        path: str | Path,
        encoding: str = "utf-8",
        color: str | Color = Color.WHITE,
        top: bool = False,
    ):
        """Add subtitle file into merger instance to merge them."""
        with open(path if type(path) is Path else Path(path), "r") as file:
            data = file.buffer.read().decode(encoding)
            subtitle = self._extract_subtitle_entries(data, color, top)
            self.subtitles.append(subtitle)

    def merge(self):
        """Merge subtitles and save output."""
        self.merged_subtitles = []
        self.timestamps = sorted(list(set(self.timestamps)))
        count = 1
        for timestamp in self.timestamps:
            for subtitle_order, subtitle in enumerate(self.subtitles, start=1):
                if timestamp in subtitle.keys():
                    # Remove \n characters at end of first subtitle texts to allow join the other subtitle
                    text = re.sub(r"[\r\n]*$", "", subtitle[timestamp])

                    # Remove Timecodes from subtitles with orders greater than 1
                    if subtitle_order > 1:
                        text = re.sub(TIMECODES_PATTERN, "", text)

                    logger.debug(f"{subtitle_order = }, text = {repr(text)}")

                    try:
                        encoded_subtitle_entry = bytes(
                            text, encoding=self.output_encoding
                        )
                    except Exception as e:
                        logger.error(
                            (
                                f'Encoding problem in "{repr(text)}" with {self.output_encoding}.'
                                f"\nError: {e}"
                            )
                        )
                        encoded_subtitle_entry = b"An error has been occurred in encoding by specified `output_encoding`"

                    if count == 1:
                        encoded_sequential_number = self._insert_bom(
                            bytes(f"{LF}{str(count)}", encoding=self.output_encoding),
                            self.output_encoding,
                        )
                    else:
                        encoded_sequential_number = LF.encode(
                            self.output_encoding
                        ) + bytes(f"{LF}{str(count)}", encoding=self.output_encoding)

                    encoded_subtitle = (
                        # sequential_number is required just before first subtitle text
                        (encoded_sequential_number if subtitle_order == 1 else b"")
                        + LF.encode(self.output_encoding)
                        + encoded_subtitle_entry
                    )
                    self.merged_subtitles.append(encoded_subtitle)
            count += 1

        if self.merged_subtitles[-1].endswith(b"\x00\n\x00"):
            self.merged_subtitles[-1] = self.merged_subtitles[-1][:-3] + b"\x00"
        if self.merged_subtitles[-1].endswith(LF_AS_BYTE):
            self.merged_subtitles[-1] = self.merged_subtitles[-1][:-1] + b""

        with open(self.output_path, "w", encoding=self.output_encoding) as output:
            output.buffer.writelines(self.merged_subtitles)
            logger.info(f'"{output.name}" created successfully.')


# How to use?

# m = Merger(output_path="en-fa.srt")
# m.add(Path('./tests/test_assets/en.srt'), color=Color.BLUE)
# m.add(Path('./tests/test_assets/fa.srt'), color=Color.YELLOW3, encoding="cp1256")
# m.merge()



#!/usr/bin/env python3

import unittest
from unittest.mock import patch

from srtmerger.main import create_parser, parse_subtitle_spec


class TestSubtitleMerger(unittest.TestCase):
    def setUp(self):
        self.parser = create_parser()

    def test_parse_subtitle_spec_basic(self):
        """Test parsing basic subtitle specification."""
        config = parse_subtitle_spec("file.srt")
        self.assertEqual(config.file_path, "file.srt")
        self.assertIsNone(config.encoding)
        self.assertIsNone(config.color)
        self.assertFalse(config.top)

    def test_parse_subtitle_spec_full(self):
        """Test parsing subtitle specification with all options."""
        config = parse_subtitle_spec("file.srt:encoding=utf-8:color=yellow:top")
        self.assertEqual(config.file_path, "file.srt")
        self.assertEqual(config.encoding, "utf-8")
        self.assertEqual(config.color, "yellow")
        self.assertTrue(config.top)

    def test_parse_subtitle_spec_partial(self):
        """Test parsing subtitle specification with some options."""
        config = parse_subtitle_spec("file.srt:color=blue:top")
        self.assertEqual(config.file_path, "file.srt")
        self.assertIsNone(config.encoding)
        self.assertEqual(config.color, "blue")
        self.assertTrue(config.top)

    def test_cli_basic(self):
        """Test basic CLI usage."""
        test_args = ["file1.srt", "file2.srt:encoding=utf-8:color=yellow"]
        with patch("sys.argv", ["srt-merger"] + test_args):
            args = self.parser.parse_args()
            self.assertEqual(args.subtitles, test_args)
            self.assertIsNone(args.default_encoding)
            self.assertIsNone(args.default_color)

    def test_cli_with_defaults(self):
        """Test CLI with default options."""
        test_args = [
            "-E",
            "utf-8",
            "-C",
            "white",
            "file1.srt:color=yellow",
            "file2.srt:top",
        ]
        with patch("sys.argv", ["srt-merger"] + test_args):
            args = self.parser.parse_args()
            self.assertEqual(len(args.subtitles), 2)
            self.assertEqual(args.default_encoding, "utf-8")
            self.assertEqual(args.default_color, "white")

    @patch("srtmerger.main.Merger")
    def test_main_integration(self, MockMerger):
        """Test main function integration."""
        mock_merger = MockMerger.return_value
        test_args = [
            "-E",
            "utf-8",
            "-C",
            "white",
            "file1.srt:color=yellow",
            "file2.srt:top",
        ]

        with patch("sys.argv", ["srtmerger"] + test_args):
            from srtmerger.main import main

            main()

            # Verify Merger was initialized
            MockMerger.assert_called_once()

            # Verify add() was called for each subtitle
            self.assertEqual(mock_merger.add.call_count, 2)

            # Verify merge() was called
            mock_merger.merge.assert_called_once()


if __name__ == "__main__":
    unittest.main(verbosity=2)



from datetime import datetime, timezone


def time_to_timestamp(time: str|datetime):
    """Turn a time string or datetime instance to timestamp in seconds or milliseconds"""
    # OSError: [Errno 22] Invalid argument
    # The timestamp() method may fail on systems like Windows for dates before
    # the UNIX epoch (1970-01-01) or far into the future.
    # logger.debug(f"⌛ {repr(time)} <- {time}, {type(time)}")
    if  type(time) is str:
        if ',' in time:            
            time, milliseconds = time.split(',')
            timestamp = (
                datetime
                .strptime(time, '%H:%M:%S')
                .replace(tzinfo=timezone.utc)
                .timestamp()
            )
            timestamp_with_milliseconds = (
                timestamp +
                (
                    (
                        int(milliseconds) / (10**len(milliseconds))
                    ) * (-1 if (timestamp < 0) else 1)
                )
            )
            return timestamp_with_milliseconds
        return (
            datetime
            .strptime(time, '%H:%M:%S')
            .replace(tzinfo=timezone.utc)
            .timestamp()
        )
    return time.replace(tzinfo=timezone.utc).timestamp()